/*
** Module   : AbstractTempTable.java
** Abstract : Abstract base class for static and dynamic temp-table 4GL objects.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20131115 Created initial version.
** 002 CA  20131209 The dynamic flag must be passed up the c'tor call chain.
** 003 HC  20131215 Implemented ADM-DATA and UNIQUE-ID attribute.
** 004 SVL 20140106 ERROR-STRING attribute and some error handling functions are now handled by
**                  this class.
** 005 SVL 20140124 Some functionality was moved here from the subclasses. Implemented
**                  hasRecords().
** 006 SVL 20140205 Got some functionality from subclasses, including COPY-TEMP-TABLE.
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;
import org.hibernate.*;
import java.util.*;

/**
 * Abstract base class for static and dynamic temp-table 4GL objects.
 */
public abstract class AbstractTempTable
extends HandleChain
implements TempTable,
           Deletable
{
   /** Default prefix for the name of a temp-table copied with COPY-TEMP-TABLE. */
   protected static final String DEFAULT_COPY_TEMP_TABLE_PREFIX = "cpy";

   /** Default buffer handle. */
   protected handle defaultBufferHandle = null;

   /** Corresponds ERROR-STRING attribute. **/
   protected final character errorString = new character();

   /** Corresponds ERROR attribute. **/
   protected final logical error = new logical(false);

   /** Corresponds ADM-DATA attribute. **/
   protected final character admData = new character();

   /** Corresponds UNIQUE-ID attribute. **/
   private final integer uniqueId = new integer();

   /**
    * Default constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or a dynamic temp-table.
    */
   public AbstractTempTable(boolean dynamic)
   {
      super(dynamic);
      uniqueId.assign(UniqueIdGenerator.getUniqueId(UniqueIdGenerator.IdKind.TEMP_TABLE));
   }

   /**
    * Gets the the unique ID number associated to this object by the
    * underlying system. Not the same as the handle value itself.
    *
    * @return   The integer number of the children.
    */
   @Override
   public integer getUniqueID()
   {
      if (uniqueId == null)
      {
         // not yet initialized, can occur at construction time
         return new integer();
      }

      return new integer(this.uniqueId);
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other)
   {
      return copyTempTable(other, new logical(false));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other, logical append)
   {
      return copyTempTable(other, append, new logical(false));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other, logical append, logical replace)
   {
      return copyTempTable(other, append, replace, new logical(false));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other, logical append, logical replace, logical looseCopy)
   {
      return copyTempTable(other, append, replace, looseCopy, new character());
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other, boolean append)
   {
      return copyTempTable(other, new logical(append));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other, boolean append, boolean replace)
   {
      return copyTempTable(other, new logical(append), new logical(replace));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other, boolean append, boolean replace, boolean looseCopy)
   {
      return copyTempTable(other,
            new logical(append),
            new logical(replace),
            new logical(looseCopy));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    * @param   prefix
    *          Prefix for naming the copied temp-table.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle  other,
                                boolean append,
                                boolean replace,
                                boolean looseCopy,
                                String  prefix)
   {
      return copyTempTable(other,
            new logical(append),
            new logical(replace),
            new logical(looseCopy),
            new character(prefix));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    * @param   prefix
    *          Prefix for naming the copied temp-table.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle    other,
                                logical   append,
                                logical   replace,
                                logical   looseCopy,
                                character prefix)
   {
      RecordBuffer otherBuf = getRecordBufferByHandle(other, CallingFunction.COPY_TEMP_TABLE);
      if (otherBuf == null)
         return new logical(false);

      if (!prepared().booleanValue())
      {
         if (addFunctionWasCalled())
         {
            displayCopyTempTableHalfDefined();
            return new logical(false);
         }

         boolean res = createTableLike(other).booleanValue();
         if (!res)
            return new logical(false);

         StringBuilder tableName = new StringBuilder(prefix.isUnknown() ?
                                                     DEFAULT_COPY_TEMP_TABLE_PREFIX :
                                                     prefix.getValue());
         tableName.append(otherBuf.getParentTable().name().getValue());
         tempTablePrepare(tableName.toString());
      }

      Temporary srcDMO = (Temporary) otherBuf.getDMOProxy();
      Temporary dstDMO = (Temporary) defaultBufferHandle().getResource();

      boolean res = TemporaryBuffer.copyAllRows(srcDMO,
                                                dstDMO,
                                                append.booleanValue() || replace.booleanValue(),
                                                true,
                                                append.booleanValue(),
                                                replace.booleanValue(),
                                                looseCopy.booleanValue());
      return new logical(res);
   }

   /**
    * Sets the ERROR-STRING attribute of 4GL temp-table.
    *
    * @param   errStr
    *          The new value for the ERROR-STRING attribute.
    */
   public void changeErrorString(character errStr)
   {
      if (!prepared().getValue())
      {
         ErrorManager.displayError(12346,
               "ERROR-STRING can only be used with a PREPARED temp-table that is not a BEFORE-TABLE",
               false);
         ErrorManager.recordOrThrowError(3131,
               "Unable to set attribute ERROR-STRING in widget of type TEMP-TABLE",
               false);
         return;
      }

      this.errorString.assign(errStr);
   }

   /**
    * Obtain the current ERROR-STRING attribute of 4GL temp-table.
    *
    * @return  the current ERROR-STRING of the temp-table
    */
   public character errorString()
   {
      return errorString;
   }

   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getADMData()
   {
      return new character(this.admData);
   }

   /**
     * Set the value of the ADM-DATA attribute.
     * 
     * @param    value
     *           The new value.
     */
   @Override
   public void setADMData(String value)
   {
      this.admData.assign(value);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(character value)
   {
      this.admData.assign(value);
   }

   /**
    * Returns the state of the ERROR attribute
    *
    * @return    Current state of the ERROR attribute.
    */
   @Override
   public logical isError()
   {
      if (!prepared().booleanValue())
      {
         displayNotPrepared();
      }

      return error;
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   @Override
   public void setError(boolean value)
   {
      setError(new logical(value));
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   @Override
   public void setError(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute ERROR on TEMP-TABLE widget",
               false);
         return;
      }

      if (!prepared().booleanValue())
      {
         displayNotPrepared();
         ErrorManager.recordOrThrowError(3131,
               "Unable to set attribute ERROR in widget of type TEMP-TABLE",
               false);
         return;
      }

      error.assign(value);
   }

   /**
    * Checks whether this temp-table contains any records.
    *
    * @return  true if this temp-table has records.
    */
   @Override
   public logical hasRecords()
   {
      Persistence persistence = PersistenceFactory.getInstance(DatabaseManager.TEMP_TABLE_DB);
      StringBuilder sb = new StringBuilder("select 1 from ");
      sb.append(defaultBuffer().getTable());
      sb.append(" where ");
      sb.append(TemporaryBuffer.MULTIPLEX_FIELD_NAME);
      sb.append(" = ");
      sb.append(defaultBuffer().getMultiplexID());
      try
      {
         ScrollableResults res = persistence.executeSQLQuery(sb.toString());
         if (res.first())
            return new logical(true);
      }
      catch (PersistenceException e)
      {
         ErrorManager.recordOrThrowError(e);
         return new logical(false);
      }

      Iterator<RecordBuffer> iter = BufferManager.get().activeBuffers(persistence, true);
      while(iter.hasNext())
      {
         RecordBuffer buffer = iter.next();
         TempTable tempTable = buffer.getParentTable();
         if (tempTable.equals(this))
            return new logical(true);
      }

      return new logical(false);
   }

   /**
    * Get default buffer for the static temp-table.
    *
    * @return default buffer for the static temp-table.
    */
   RecordBuffer defaultBuffer()
   {
      return RecordBuffer.get((DataModelObject) defaultBufferHandle.getResource());
   }

   /**
    * Determine if an ADD-* function was called for the table (CLEAR function resets the state).
    * Makes sense only for unprepared tables.
    *
    * @return <code>true</code> if an ADD-* function was called for the table.
    */
   protected abstract boolean addFunctionWasCalled();

   /**
    * Display "CREATE-LIKE must be the first method on a newly CREATEd TEMP-TABLE, or a newly
    * CLEARed one" error message.
    */
   protected void displayCreateLikeNotFirst()
   {
      ErrorManager.displayError(9041,
            "CREATE-LIKE must be the first method on a newly CREATEd TEMP-TABLE, " +
                  "or a newly CLEARed one",
            false);
   }

   /**
    * Display "Definitional TEMP-TABLE methods (ADD-NEW/LIKE) may not be called after
    * TEMP-TABLE-PREPARE has been called unless a CLEAR is done first, and may never be called on
    * a DEFINED static table" error message.
    */
   protected void displayPrepared()
   {
      ErrorManager.displayError(9042,
            "Definitional TEMP-TABLE methods (ADD-NEW/LIKE) may not be called after " +
                  "TEMP-TABLE-PREPARE has been called unless a CLEAR is done first, and may never " +
                  "be called on a DEFINED static table",
            false);

   }

   /**
    * Display "TEMP-TABLE is already prepared--ignoring" error message.
    */
   protected void displayPreparedIgnoring()
   {
      ErrorManager.displayError(9043, "TEMP-TABLE is already prepared--ignoring", false);
   }

   /**
    * Display "Unable to assign UNKNOWN value to attribute UNDO on TEMP-TABLE widget" error
    * message.
    */
   protected void displayUnableAssignUnknownUndo()
   {
      ErrorManager.displayError(4083,
            "**Unable to assign UNKNOWN value to attribute UNDO on TEMP-TABLE widget",
            false);
   }

   /**
    * Display "CLEAR method may not be called on a static temp-table object, nor on a dynamic one
    * with records; use buffer:EMPTY-TEMP-TABLE" error message.
    */
   protected void displayCannotClear()
   {
      ErrorManager.displayError(9085,
            "CLEAR method may not be called on a static temp-table object, nor on a dynamic " +
                  "one with records; use buffer:EMPTY-TEMP-TABLE",
            false);
   }

   /**
    * Display "TEMP-TABLE must be PREPARED before referencing NAME/DEFAULT-BUFFER-HANDLE etc"
    * error message.
    */
   protected void displayNotPrepared()
   {
      ErrorManager.displayError(9040,
            "TEMP-TABLE must be PREPARED before referencing NAME/DEFAULT-BUFFER-HANDLE etc",
            false);

   }

   /**
    * Display "Unable to CREATE-LIKE for TEMP-TABLE--try using a smaller source table" error
    * message.
    */
   protected void displayUnableCreateLike()
   {
      ErrorManager.displayError(9047,
            "Unable to CREATE-LIKE for TEMP-TABLE--try using a smaller source table",
            false);
   }

   /**
    * Display "Unable to ADD-FIELDS-FROM for TEMP-TABLE" error message.
    */
   protected void displayUnableAddFieldsFrom()
   {
      ErrorManager.displayError(9049, "Unable to ADD-FIELDS-FROM for TEMP-TABLE", false);
   }

   /**
    * Display "TEMP-TABLE method could not find LIKE table from BUFFER handle" error message.
    */
   protected void displayCouldNotFindLikeTableFromHandle()
   {
      ErrorManager.displayError(9058,
            "TEMP-TABLE method could not find LIKE table from BUFFER handle",
            false);
   }

   /**
    * Display "COPY-TEMP-TABLE must have valid buffer/temp-table handle as first argument" error
    * message.
    */
   protected void displayCopyTempTableInvalidFirstArg()
   {
      ErrorManager.displayError(12339,
            "COPY-TEMP-TABLE must have valid buffer/temp-table handle as first argument",
            false);
   }

   /**
    * Display "COPY-TEMP-TABLE target temp-table must be either PREPARED or CLEAR, but not
    * half-defined" error message.
    */
   protected void displayCopyTempTableHalfDefined()
   {
      ErrorManager.displayError(12341,
       "COPY-TEMP-TABLE target temp-table must be either PREPARED or CLEAR, but not half-defined",
       false);
   }

   /**
    * Get record buffer by buffer or temp-table handle. Displays error if the handle is not valid
    * or of a wrong type.
    *
    * @param  hbuf
    *         Buffer or temp-table handle.
    * @param  callingFunction
    *         Calling function. Affects error messaging.
    *
    * @return target record buffer or <code>null</code> if the case of error.
    */
   protected RecordBuffer getRecordBufferByHandle(handle hbuf, CallingFunction callingFunction)
   {
      if (!hbuf._isValid())
      {
         switch (callingFunction)
         {
            case COPY_TEMP_TABLE:
               displayCopyTempTableInvalidFirstArg();
               break;
            case ADD_FIELDS_FROM:
               displayUnableAddFieldsFrom();
               break;
            case CREATE_LIKE:
               displayUnableCreateLike();
               break;
         }

         return null;
      }

      Object handleValue = hbuf.getResource();
      Class handleClass = handleValue.getClass();
      Buffer buffer;

      if (Buffer.class.isAssignableFrom(handleClass))
      {
         buffer = (Buffer) handleValue;
      }
      else if (TempTable.class.isAssignableFrom(handleClass))
      {
         TempTable tempTable = (TempTable) handleValue;
         handle defaultBufferHandle = tempTable.prepared().booleanValue() ?
               tempTable.defaultBufferHandle() :
               null;
         if (defaultBufferHandle == null)
         {
            switch (callingFunction)
            {
               case COPY_TEMP_TABLE:
                  displayCopyTempTableInvalidFirstArg();
                  break;
               case ADD_FIELDS_FROM:
                  displayCouldNotFindLikeTableFromHandle();
                  displayUnableAddFieldsFrom();
                  break;
               case CREATE_LIKE:
                  displayCouldNotFindLikeTableFromHandle();
                  displayUnableCreateLike();
                  break;
            }
            return null;
         }

         buffer = (Buffer) defaultBufferHandle.getResource();
      }
      else
      {
         switch (callingFunction)
         {
            case COPY_TEMP_TABLE:
               displayCopyTempTableInvalidFirstArg();
               break;
            case ADD_FIELDS_FROM:
               displayCouldNotFindLikeTableFromHandle();
               displayUnableAddFieldsFrom();
               break;
            case CREATE_LIKE:
               displayCouldNotFindLikeTableFromHandle();
               displayUnableCreateLike();
               break;
         }
         return null;
      }

      return ((BufferReference) buffer).buffer();
   }

   /**
    * Enumeration which represents some temp-table functions. Required because error handing may
    * vary depending on the calling function.
    */
   protected enum CallingFunction
   {
      COPY_TEMP_TABLE,
      ADD_FIELDS_FROM,
      CREATE_LIKE
   }
}
